"""
This provides a canonical binary serialization for the main Python primitive types.
The primary use-case is to be able to produce a deterministic checksum/signature of an arbitrary data structure,
but deserialization is implemented as well.
For faster, space-efficient and wide compatibility serialization where deterministic form is not needed, use msgpack.
"""
import hashlib
import re
import struct
from io import BytesIO
from typing import Type, Callable, Optional, Union, Any

from tarka.utility.algorithm.traverse import traverse, TraverseType, SORTED_STR_TRAVERSE, TraverseTypeMap

StreamProcessFn = Callable[[bytes], Any]
TraverseDeterministicTransformMap = dict[Type[TraverseType], Callable[[StreamProcessFn, TraverseType], None]]


def _ppd_int(acc: StreamProcessFn, o: int):
    b = str(o).encode("utf-8")
    acc(str(len(b)).encode("utf-8"))
    acc(b"i")
    acc(b)


def _ppd_str(acc: StreamProcessFn, o: str):
    b = o.encode("utf-8")
    acc(str(len(b)).encode("utf-8"))
    acc(b"s")
    acc(b)


def _ppd_float(acc: StreamProcessFn, o: float):
    acc(b"f")
    acc(struct.pack(">d", o))


def _ppd_bool(acc: StreamProcessFn, o: bool):
    acc(b"T" if o else b"F")


def _ppd_none_type(acc: StreamProcessFn, o):
    acc(b"N")


def _ppd_bytes(acc: StreamProcessFn, o: bytes):
    acc(str(len(o)).encode("utf-8"))
    acc(b"b")
    acc(o)


def _ppd_list(acc: StreamProcessFn, o: list):
    acc(str(len(o)).encode("utf-8"))
    acc(b"l")


def _ppd_dict(acc: StreamProcessFn, o: dict):
    acc(str(len(o) << 1).encode("utf-8"))
    acc(b"d")


def _ppd_tuple(acc: StreamProcessFn, o: tuple):
    acc(str(len(o)).encode("utf-8"))
    acc(b"t")


def _ppd_set(acc: StreamProcessFn, o: set):
    acc(str(len(o)).encode("utf-8"))
    acc(b"h")


def _ppd_frozenset(acc: StreamProcessFn, o: frozenset):
    acc(str(len(o)).encode("utf-8"))
    acc(b"H")


VERSION_TAG = b"1v"
PYTHON_PRIMITIVES_DETERMINISTIC_TRANSFORM_MAP: TraverseDeterministicTransformMap = {
    str: _ppd_str,
    int: _ppd_int,
    float: _ppd_float,
    bool: _ppd_bool,
    type(None): _ppd_none_type,
    bytes: _ppd_bytes,
    list: _ppd_list,
    dict: _ppd_dict,
    tuple: _ppd_tuple,
    set: _ppd_set,
    frozenset: _ppd_frozenset,
}


def deterministic_dumps(arg: TraverseType) -> bytes:
    """
    Get the canonical binary serialization of the object in whole.
    """
    buf = BytesIO()
    deterministic_transform(arg, buf.write)
    return buf.getvalue()


def deterministic_hash(arg: TraverseType, algo: str, **algo_kw) -> bytes:
    """
    Calculate the canonical hash of the object.
    """
    h = getattr(hashlib, algo)(**algo_kw)
    deterministic_transform(arg, h.update)
    return h.digest()


def deterministic_transform(arg: TraverseType, acc: StreamProcessFn) -> None:
    """
    Process the canonical binary serialization stream of the object.
    """
    _deterministic_transform(arg, acc, prefix=VERSION_TAG)


def _deterministic_transform(
    arg: TraverseType,
    acc: StreamProcessFn,
    tr_map: TraverseDeterministicTransformMap = None,
    traverse_map: TraverseTypeMap = None,
    prefix: bytes = None,
) -> None:
    if tr_map is None:
        tr_map = PYTHON_PRIMITIVES_DETERMINISTIC_TRANSFORM_MAP
    if traverse_map is None:
        traverse_map = SORTED_STR_TRAVERSE
    if prefix is not None:
        acc(prefix)
    for x in traverse(arg, traverse_map):
        tr_map[type(x)](acc, x)


TraverseDeterministicTransformUnmap = dict[bytes, tuple[int, Callable[[bytes], TraverseType]]]

DYNAMIC_LENGTH_VALUE = -5
DYNAMIC_LENGTH_COLLECTION = -10
PYTHON_PRIMITIVES_DETERMINISTIC_TRANSFORM_UNMAP: TraverseDeterministicTransformUnmap = {
    b"s": (DYNAMIC_LENGTH_VALUE, lambda v: v.decode("utf-8")),
    b"i": (DYNAMIC_LENGTH_VALUE, lambda v: int(v.decode("utf-8"))),
    b"f": (8, lambda v: struct.unpack(">d", v)[0]),
    b"T": (0, lambda _: True),
    b"F": (0, lambda _: False),
    b"N": (0, lambda _: None),
    b"b": (DYNAMIC_LENGTH_VALUE, lambda v: v),
    b"l": (DYNAMIC_LENGTH_COLLECTION, lambda l: l),
    b"d": (DYNAMIC_LENGTH_COLLECTION, lambda l: {l[i]: l[i + 1] for i in range(0, len(l), 2)}),
    b"t": (DYNAMIC_LENGTH_COLLECTION, lambda l: tuple(l)),
    b"h": (DYNAMIC_LENGTH_COLLECTION, lambda l: set(l)),
    b"H": (DYNAMIC_LENGTH_COLLECTION, lambda l: frozenset(l)),
}

LOAD_NUM_RE = re.compile(rb"([0-9]+)")


class _LoadParent:
    __slots__ = ("parent_index", "children", "child")

    def __init__(self, parent_index: int, children: int):
        self.parent_index = parent_index
        self.children = children
        self.child = 0

    def next_child(self) -> tuple[int, int, bool]:
        c = self.child
        assert c < self.children
        self.child = c + 1
        return self.parent_index, c, self.child == self.children


def deterministic_loads(b: bytes) -> TraverseType:
    """
    Restore the object from the canonical binary serialization.
    Implemented without recursion.
    """
    if not b.startswith(VERSION_TAG):
        raise ValueError()
    # List of Tuple[parent-index, child-index, collection--type-value-fn, collection-length-accum OR self-value]
    data_map: list[tuple[int, int, Optional[Callable], Union[list[TraverseType], TraverseType]]] = []
    parent_stack: list[_LoadParent] = [_LoadParent(-1, 1)]
    i = len(VERSION_TAG)
    while i < len(b):
        nm = LOAD_NUM_RE.match(b, i)
        if nm:
            g = nm.group(1)
            num = int(g.decode("utf-8"))
            i += len(g)
        else:
            num = None
        t = b[i : i + 1]
        len_type, val_fn = PYTHON_PRIMITIVES_DETERMINISTIC_TRANSFORM_UNMAP[t]
        if (num is None) is not (len_type >= 0):
            ValueError()
        if num is None:
            num = len_type
        i += 1
        pi, nc, pd = parent_stack[-1].next_child()
        if pd:
            try:
                parent_stack.pop()
            except IndexError:
                raise ValueError()
        if len_type == DYNAMIC_LENGTH_COLLECTION:
            if num > 0:
                parent_stack.append(_LoadParent(len(data_map), num))
            data_map.append((pi, nc, val_fn, [None] * num))
        else:
            data_map.append((pi, nc, None, val_fn(b[i : i + num])))
            i += num
    if parent_stack:
        # serialized data did not end correctly
        raise ValueError()
    for i in range(len(data_map) - 1, 0, -1):
        parent_index, child_index, collection_type_value_fn, value = data_map[i]
        data_map[parent_index][3][child_index] = (
            value if collection_type_value_fn is None else collection_type_value_fn(value)
        )
    _, _, collection_type_value_fn, value = data_map[0]
    return value if collection_type_value_fn is None else collection_type_value_fn(value)



import asyncio
import sqlite3
import time
from contextlib import contextmanager, asynccontextmanager
from functools import partial
import queue
from typing import Callable, Optional, Any, Sequence, TypeVar, Type, AsyncContextManager

import wait_for2

from tarka.utility.thread import AbstractThread


def _callback_result(future: asyncio.Future, result):
    if not future.done():
        future.set_result(result)


def _callback_exception(future: asyncio.Future, exc):
    if not future.done():
        future.set_exception(exc)


def sqlite_retry(
    fn: Callable[[], Any],
    retry_timeout: Optional[float] = None,
    wait_time: float = 0.001,
    max_wait_time: float = 0.15,
    wait_multiplier: float = 1.5,
) -> Any:
    """
    Simple retry logic for handling SQLite operational errors when necessary. Keep in mind that using this can and
    should be avoided usually, because the builtin busy_timeout handler is set up by default. There are a few
    exceptions that do not utilize the busy_handler even if it is set, like the wal_checkpoint pragma.

    The wait time can be absolute (to the process by time.perf_counter()) to retry until, or a duration to retry for.
    The duration can be expressed by negative values, which will be translated into absolute time when the first error
    is raised.
    """
    while True:
        try:
            return fn()
        except sqlite3.OperationalError:
            if not retry_timeout:
                raise
            elif retry_timeout < 0:
                retry_timeout = time.perf_counter() - retry_timeout
            elif time.perf_counter() > retry_timeout:
                raise
        time.sleep(wait_time)
        # adjust wait time to backoff next time if error persists
        if wait_time < max_wait_time:
            wait_time *= wait_multiplier


_ClsT = TypeVar("_ClsT")


class AbstractAioSQLiteDatabase(AbstractThread):
    """
    Provide a lightweight asyncio compatible, customizable interface to an arbitrary SQLite database in a safe way.
    All SQL connection operations are restricted to be executed on the worker thread, guaranteeing serialization
    requirements. If more processes would access the database the transaction_mode selector can be used, but job
    specific transaction handling would be needed for optimal performance.

    Requests shall be implemented like this:

        def _get_all_impl(self):
            return self._con.execute("SELECT x FROM y").fetchall()

        get_all = partialmethod(AbstractAioSQLiteDatabase._run_job, _get_all_impl)

    """

    __slots__ = ("_loop", "_con", "_request_queue", "_timeout", "started", "closed")

    @classmethod
    @asynccontextmanager
    async def create(cls: Type[_ClsT], *args, **kwargs) -> AsyncContextManager[_ClsT]:
        """
        Strict start-stop helper for automatic cleanup.
        """
        self = cls(*args, **kwargs)
        try:
            self.start()
            await self.wait_ready()
            yield self
        finally:
            self.stop()

    def __init__(self, sqlite_db_path: str, sqlite_timeout: float = 60.0):
        self._loop = asyncio.get_running_loop()
        self._con: sqlite3.Connection = None
        self._request_queue: queue.Queue = None
        self._timeout = sqlite_timeout
        AbstractThread.__init__(self, (sqlite_db_path,))
        self.started = asyncio.Event()
        self.closed = asyncio.Event()

    async def wait_ready(self):
        if self.is_alive():  # only wait for start/stop if the worker thread has been started
            started_task = self._loop.create_task(self.started.wait())
            closed_task = self._loop.create_task(self.closed.wait())
            _, pending = await asyncio.wait([started_task, closed_task], return_when=asyncio.FIRST_COMPLETED)
            if pending:  # cleanup tasks to avoid them being logged as abandoned
                for t in pending:
                    t.cancel()
                await asyncio.gather(*pending, return_exceptions=True)
        if self.closed.is_set() or not self.started.is_set():
            raise Exception(f"SQLite connection could not start for {self.__class__.__name__}")

    def start(
        self,
        args: Optional[Sequence[Any]] = None,
        callback: Optional[Callable[[], None]] = None,
        daemon: Optional[bool] = None,
        name_prefix: Optional[str] = None,
    ):
        if args is not None:  # pragma: no cover
            raise ValueError("SQLite worker does not support custom args at start.")
        if callback is not None:  # pragma: no cover
            raise ValueError("SQLite worker does not support custom callback. Use the 'closed' event.")
        if daemon not in (None, False):  # pragma: no cover
            raise ValueError("SQLite worker must not be daemon to ensure cleanup.")
        self._request_queue = queue.Queue()
        AbstractThread.start(self, None, partial(self._loop.call_soon_threadsafe, self.closed.set), False, name_prefix)

    def stop(self, timeout: Optional[float] = 0) -> bool:
        """
        The default timeout is zero, we assume the .closed event will be used to wait for cleanup in asyncio.
        """
        q = self._request_queue  # saving a reference resolves race-conditions with worker thread
        if q is not None:
            q.put_nowait(None)
        return AbstractThread.stop(self, timeout)

    def _thread(self, db_path: str) -> None:
        """
        SQLite connection and request executor thread.
        """
        self._con = sqlite3.connect(db_path, timeout=self._timeout, isolation_level=None)
        try:
            # initialise the db
            self._initialise()
            with self._transact(begin_immediate=True):
                self._setup()
            self._loop.call_soon_threadsafe(self.started.set)
            # run job queue
            while True:
                job = self._request_queue.get()
                if job is None:
                    break
                process_fn, args, post_process_fn, aio_future, begin_immediate = job
                if aio_future.done():  # Could have been cancelled before we got to it.
                    continue
                try:
                    with self._transact(begin_immediate):
                        result = process_fn(self, *args)
                    if post_process_fn:
                        post_process_fn(self)
                    self._loop.call_soon_threadsafe(_callback_result, aio_future, result)
                except Exception as e:
                    self._loop.call_soon_threadsafe(_callback_exception, aio_future, e)
        finally:
            # prevent more jobs to be queued
            q = self._request_queue
            self._request_queue = None
            # notify dead jobs if any
            try:
                while True:
                    job = q.get_nowait()
                    if job:
                        job[3].cancel()
            except queue.Empty:
                pass
            # close the database
            con = self._con
            self._con = None
            con.close()

    @contextmanager
    def _transact(self, begin_immediate: bool):
        if begin_immediate:
            self._con.execute("BEGIN IMMEDIATE")
        else:
            self._con.execute("BEGIN")
        try:
            yield
        except BaseException:
            self._con.execute("ROLLBACK")
            raise
        else:
            self._con.execute("COMMIT")

    async def _run_job(
        self,
        process_fn: Callable,
        *args,
        post_process_fn: Optional[Callable] = None,
        begin_immediate: bool = True,
        timeout: Optional[float] = None,
    ):
        """
        This is designed to be used as

            get_xy = partialmethod(AbstractAioSQLiteDatabase._run_job, _get_xy_impl)

        which means the wrapped function is not yet bound in the expression. The self argument is automatically
        passed by the worker thread to make it work.
        This will raise AttributeError if the database has been closed.
        """
        f = self._loop.create_future()
        self._request_queue.put_nowait((process_fn, args, post_process_fn, f, begin_immediate))
        return await wait_for2.wait_for(f, timeout)

    def _initialise(self):
        """
        The connection is ready and the database initialization like pragma definitions shall be done here.

        By default the db will use WAL journaling and NORMAL sync policy. These supply the highest throughput
        with persistence, when the application can choose to ensure durability at any point by issuing a checkpoint.
        Additionally the "wal_autocheckpoint" pragma could be tuned depending on the database use-pattern.

        The "busy_timeout" pragma does not need to be set here, as the python binding of SQLite3 does that in the
        connection, for which it uses the "timeout" argument.

        NOTE: Current execution is not inside a transaction!
        """
        self._con.execute("PRAGMA journal_mode = WAL;")
        self._con.execute("PRAGMA synchronous = NORMAL;")

    def _setup(self):
        """
        The connection is ready and the database initialization like schema shall be done.
        """
        raise NotImplementedError()

    def _checkpoint(self):
        """
        Use in WAL mode with NORMAL synchronous operation to ensure writes are synced to storage!
        When guaranteed durability is required at a point, this can be used as a post-process callback like:

            mut_xy = partialmethod(AbstractAioSQLiteDatabase._run_job, _mut_xy_impl, post_process_fn=_checkpoint)

        If there are multiple writer processes the checkpoint api call can easily return SQLITE_BUSY.
        If the database is used by a single process in a single instance, such will not happen.
        """
        if self._con.execute("PRAGMA wal_checkpoint(TRUNCATE);").fetchall()[0][0] != 0:
            raise sqlite3.OperationalError("WAL checkpoint failed: SQLITE_BUSY")

    def _checkpoint_retrying(self):
        """
        To be used similarly to ._checkpoint(), but this will implicitly retry similar to normal operations as
        configured initial timeout value.
        Use if multiple connections are expected to operate on the DB and one will require explicit checkpoints
        at specific places this may work well enough.
        Keep in mind though, that if more durability is required, the synchronous pragma could be tuned instead.
        """
        sqlite_retry(self._checkpoint, retry_timeout=-self._timeout)



from threading import Thread
from typing import Optional, Sequence, Any, Callable


class AbstractThread(object):
    """
    Simple thread boilerplate or wrapper with support for stopping by request and a callback at exit.
    """

    __slots__ = ("__args", "__alive", "__t")

    def __init__(self, args: Optional[Sequence[Any]] = None):
        self.__args = args
        self.__alive: bool = True
        self.__t: Optional[Thread] = None

    def is_alive(self):
        """
        Determine if the thread is running and not terminating.
        """
        return self.__t is not None and self.__alive and self.__t.is_alive()

    def is_finished(self):
        """
        Determine if the thread has started and is still running.
        """
        return self.__t is not None and not self.__t.is_alive()

    def _keep_running(self):
        """
        The thread should poll this periodically to check if it has to terminate.
        """
        return self.__alive

    def start(
        self,
        args: Optional[Sequence[Any]] = None,
        callback: Optional[Callable[[], None]] = None,
        daemon: Optional[bool] = None,
        name_prefix: Optional[str] = None,
    ) -> None:
        if self.__t is not None or not self.__alive:
            raise Exception(f"Restarting {self.__class__.__name__} is not supported")
        if args is None:
            if self.__args is None:
                args = ()
            else:
                args = self.__args
        elif self.__args:
            raise Exception("Only provide thread arguments in __init__() or start()")
        self.__t = Thread(
            target=self._thread_wrapper,
            args=(callback, args),
            daemon=daemon,
            name=f"{name_prefix}-{self.__class__.__name__}" if name_prefix else self.__class__.__name__,
        )
        del self.__args  # help gc
        self.__t.start()

    def stop(self, timeout: Optional[float] = None) -> bool:
        """
        Request the thread to stop and wait for it if desired. Returns True if the thread has stopped before returning.
        """
        self.__alive = False
        if (timeout is None or timeout > 0) and not self.is_finished():
            self.__t.join(timeout)
        return self.is_finished()

    def _thread_wrapper(self, callback: Optional[Callable[[], None]], args: Sequence[Any]) -> None:
        """
        Provides a finish-callback functionality independent of the main thread execution.
        """
        try:
            self._thread(*args)
        finally:
            if callback:
                callback()

    def _thread(self, *args: Any) -> None:
        """
        Main thread execution must be implemented here.
        The args were received either from the __init__() or the start() call.
        """
        raise NotImplementedError()



from datetime import datetime
from typing import Union

from tarka.utility.time.utc import utc_datetime_from_timestamp


class PrettyTimestamp:
    """
    Timestamps that are human readable, concise, constant-width, with no special characters by default.
    Timezone information is omitted. The same tz should be used in the same context to avoid ambiguity.
    """

    __slots__ = ("_date", "_time", "_date_time", "_date_time_ms")

    def __init__(self, date: str = "%y%m%d", time: str = "%H%M%S", datetime_sep: str = "", ms_sep: str = ""):
        self._date = date
        self._time = time
        self._date_time = f"{date}{datetime_sep}{time}"
        self._date_time_ms = f"{date}{datetime_sep}{time}{ms_sep}%f"

    def date(self, dt: Union[datetime, float, int]) -> str:
        if not isinstance(dt, datetime):
            dt = utc_datetime_from_timestamp(dt)
        return dt.strftime(self._date)

    def time(self, dt: Union[datetime, float, int]) -> str:
        if not isinstance(dt, datetime):
            dt = utc_datetime_from_timestamp(dt)
        return dt.strftime(self._time)

    def date_time(self, dt: Union[datetime, float, int]) -> str:
        if not isinstance(dt, datetime):
            dt = utc_datetime_from_timestamp(dt)
        return dt.strftime(self._date_time)

    def date_time_milli(self, dt: Union[datetime, float, int]) -> str:
        if not isinstance(dt, datetime):
            dt = utc_datetime_from_timestamp(dt)
        return dt.strftime(self._date_time_ms)[:-3]

    def date_time_micro(self, dt: Union[datetime, float, int]) -> str:
        if not isinstance(dt, datetime):
            dt = utc_datetime_from_timestamp(dt)
        return dt.strftime(self._date_time_ms)


PRETTY_TIMESTAMP = PrettyTimestamp()  # default global instance for ease of use



from typing import Tuple


def split_extension(name: str) -> Tuple[str, str]:
    """
    Split extension that respects the common, but peculiar exception of multi-level extensions like ".tar.gz".
    Since the extension of name is pretty much optional, arbitrary and has no real standard, the only way to deal
    with them is to care for each specific case we are interested in. Therefore this solution is never going to be
    complete, but at least it is a solution to some.

    NOTE: Additional differences to os.path.splitext():
        - input can't be a path, only a filename is correct
        - ending periods will not be split as an extension
        - no support for bytes

    One can argue that "there is no actual multi-level extension and .tar.gz is just .gz". There is no pertinent
    case when I want to create an archive "foo.tar.gz" and I consider "foo.tar" the name, so the ".tar" must be
    part of the extension by exclusion. The fact that this question arises at all, points out the absolute lack of
    any guidance on what an extension should actually be. It is and always was a form of metadata that indicates
    the type of content the file holds. The most notable system where this convention is kept is Windows, but even
    for them, the official docs only state that an extension is "optional" and "separated from the filename by a
    period": https://learn.microsoft.com/en-us/windows/win32/fileio/naming-a-file

    NOTE: Validating the name for use is not in the scope of this function to be usable by the OS or FS.
    """
    rightmost_dot = name.rfind(".")
    if rightmost_dot <= 0 or rightmost_dot == len(name) - 1:
        # Either no dot, leading single dot or trailing dot. All of which means that there is no extension to split.
        return name, ""
    for after_hidden in range(rightmost_dot):
        if name[after_hidden] != ".":
            # Found first not-dot that is before the rightmost dot.
            break
    else:
        # All dots are leading, meaning there is no extension to split.
        return name, ""
    second_rightmost_dot = name.rfind(".", after_hidden, rightmost_dot)
    if second_rightmost_dot > after_hidden and 7 <= len(name) - second_rightmost_dot <= 8:
        # There is a second rightmost dot that is right of the first not-dot and it is in the range from the end of
        # the name to be a nested-extension possibly.
        ext = name[second_rightmost_dot:]
        if ext.lower() in (".tar.gz", ".tar.bz2", ".tar.xz"):
            return name[:second_rightmost_dot], ext
    return name[:rightmost_dot], name[rightmost_dot:]



from __future__ import annotations

import logging
import logging.handlers
from datetime import datetime
from typing import Union

LOGGER_FORMAT_VERBOSE = "%(asctime)s|%(levelname).1s|%(name)s:%(funcName)s:%(lineno)d| %(message)s"
LOGGER_FORMAT_VERBOSE_THREAD = "%(asctime)s|%(thread)d|%(levelname).1s|%(name)s:%(funcName)s:%(lineno)d| %(message)s"
LOGGER_FORMAT_TIME_CONCISE = "%y%m%d %H%M%S"
LOGGER_FORMAT_MSEC_WITH_DOT = "%s.%03d"  # use decimal dot


def create_logging_formatter(
    fmt: str = None,
    time_format: str = None,
    msec_format: str = None,
    concise_asctime: bool = False,
    include_thread: bool = False,
) -> logging.Formatter:
    """
    Better default formats and support for overriding the formatting of asctime, which has no API. (datefmt lacks msec)
    """
    if fmt:
        pass
    elif include_thread:
        fmt = LOGGER_FORMAT_VERBOSE_THREAD
    else:
        fmt = LOGGER_FORMAT_VERBOSE
    formatter = logging.Formatter(fmt)
    if formatter.usesTime():  # noqa: this is not publicly documented... should be avaialable since Python 3.2
        if time_format:
            pass
        elif concise_asctime:
            time_format = LOGGER_FORMAT_TIME_CONCISE
        else:
            time_format = logging.Formatter.default_time_format
        if msec_format:
            pass
        else:
            msec_format = LOGGER_FORMAT_MSEC_WITH_DOT
        msec_format % (datetime.now().strftime(time_format), 123456)  # validate
        formatter.default_time_format = time_format
        formatter.default_msec_format = msec_format
    return formatter


class HandlerFormatterConfig:
    def __init__(
        self,
        fmt: str = None,
        time_format: str = None,
        msec_format: str = None,
        concise_asctime: bool = False,
        include_thread: bool = False,
    ):
        self.fmt = fmt
        self.time_format = time_format
        self.msec_format = msec_format
        self.concise_asctime = concise_asctime
        self.include_thread = include_thread

    @property
    def formatter(self) -> logging.Formatter:
        return create_logging_formatter(
            self.fmt, self.time_format, self.msec_format, self.concise_asctime, self.include_thread
        )

    @classmethod
    def apply(cls, handler: logging.Handler, formatter: HandlerFormatterOrHandlerFormatterConfig) -> None:
        if isinstance(formatter, HandlerFormatterConfig):
            formatter = formatter.formatter
        handler.setFormatter(formatter)


HandlerFormatterOrHandlerFormatterConfig = Union[logging.Formatter, HandlerFormatterConfig]



import asyncio
from contextlib import asynccontextmanager
from typing import Type, Dict, Any

from alembic.config import Config
from sqlalchemy import event
from sqlalchemy.exc import DatabaseError
from sqlalchemy.ext.asyncio import create_async_engine, AsyncConnection, AsyncSession, AsyncEngine
from sqlalchemy.orm import sessionmaker

from tarka.asqla.alembic import get_alembic_config, AlembicHelper
from tarka.asqla.tx import (
    TransactionExecutor,
    RetryableTransactionFactory,
    SessionTransactionExecutor,
    RetryableSessionTransactionFactory,
)


class Database(object):
    engine: AsyncEngine = None
    serializable_engine: AsyncEngine = None
    session_maker: Type[AsyncSession] = None

    def __init__(
        self,
        alembic_dir: str,
        connect_url: str,
        engine_kwargs: Dict[str, Any] = None,  # like echo, connect_args, etc.
        session_maker_kwargs: Dict[str, Any] = None,  # like expire_on_commit, autoflush, etc.
        aiosqlite_serializable_begin: str = "BEGIN",  # serializable tx support workaround for aiosqlite
    ):
        self.alembic_dir = alembic_dir
        self._connect_url = connect_url
        self._engine_kwargs = engine_kwargs
        self._session_maker_kwargs = session_maker_kwargs
        self._aiosqlite_serializable_begin = aiosqlite_serializable_begin
        self.alembic_head_at_startup = ""

    def get_alembic_config(self) -> Config:
        return get_alembic_config(self.alembic_dir, str(self.engine.sync_engine.url))

    async def _init_engine(self):
        """
        If further customization is necessary for the engines, this shall be overridden.
        To switch SQLite to WAL mode for example:

            if self.engine.dialect.name == "sqlite":
                @event.listens_for(self.engine.sync_engine, "connect")
                def _setup_sqlite_connection(dbapi_con, con_record):
                    dbapi_con.execute("PRAGMA journal_mode = WAL;")
                    dbapi_con.execute("PRAGMA synchronous = NORMAL;")
        """
        self.engine = create_async_engine(self._connect_url, **(self._engine_kwargs or {}))
        self.serializable_engine = self.engine.execution_options(isolation_level="SERIALIZABLE")
        if self.engine.driver == "aiosqlite" and self._aiosqlite_serializable_begin:
            # aiosqlite (<=0.20) does not honor the isolation_level with appropriate BEGIN commands, in fact it
            # never issues begin commands, transactions always start implicitly for reads.
            # SQLite itself does not support concurrent mutating transactions in any mode. In WAL journal_mode, there
            # can be parallel and concurrent readers, but writing is always with an exclusive lock. In this sense "any"
            # transaction will actually be serializable provided that an explicit BEGIN is emitted. The actual
            # difference is the point and time that parallel connections encounter locking errors.
            # The workaround (if not disabled) is to explicitly execute a BEGIN command at transaction start.
            # The command can be set for the database (DEFERRED, IMMEDIATE or EXCLUSIVE) when relevant.
            @event.listens_for(self.serializable_engine.sync_engine, "begin")
            def do_begin(conn):
                conn.exec_driver_sql(self._aiosqlite_serializable_begin)

        self.session_maker = sessionmaker(self.engine, class_=AsyncSession, **(self._session_maker_kwargs or {}))

    async def startup(self):
        """
        Initialize engine first, then bootstrap or migrate schema of the database to be up-to-date.
        """
        await self._init_engine()
        alembic_helper = AlembicHelper(self.get_alembic_config())
        retry = 0
        while True:  # handle conflicting migration attempt by parallel workers
            try:
                async with self.engine.begin() as connection:
                    connection: AsyncConnection
                    self.alembic_head_at_startup = await connection.run_sync(alembic_helper.run_strip_output, "current")

                    await self._upgrade(alembic_helper, connection)
                    await self._start_hook(connection)
            except DatabaseError:
                if retry >= 5:
                    raise
                retry += 1
                await asyncio.sleep(0.25)
            else:
                break

    async def _upgrade(self, alembic_helper: AlembicHelper, connection: AsyncConnection):
        if not self.alembic_head_at_startup.endswith(" (head)"):
            await self._pre_upgrade_hook(connection)
            await connection.run_sync(alembic_helper.run, "upgrade", "head")
            await self._post_upgrade_hook(connection)

    async def _pre_upgrade_hook(self, connection: AsyncConnection):
        """
        Place to put custom DDL commands before alembic upgrade.
        Keep in mind that it is usually better to implement all changes as alembic revisions.
        NOTE: This is called before bootstrap (first upgrade) as well.
        """

    async def _post_upgrade_hook(self, connection: AsyncConnection):
        """
        Place to put custom DDL commands after alembic upgrade.
        Keep in mind that it is usually better to implement all changes as alembic revisions.
        """

    async def _start_hook(self, connection: AsyncConnection):
        """
        Place to put custom DDL commands that are executed each time a server starts.
        The connection is currently at the HEAD alembic revision, in the upgrade transaction.
        """

    async def shutdown(self):
        await self.engine.dispose()

    @asynccontextmanager
    async def run(self):
        """
        Convenience wrapper for startup & shutdown.
        """
        await self.startup()
        try:
            yield self
        finally:
            await self.shutdown()

    def serializable_tx(self, tx_factory: RetryableTransactionFactory) -> TransactionExecutor:
        """
        Get a transaction executor with SERIALIZABLE isolation_level and automatic retry.
        """
        return TransactionExecutor(self.serializable_engine, tx_factory)

    def session_serializable_tx(self, tx_factory: RetryableSessionTransactionFactory) -> SessionTransactionExecutor:
        """
        Get an ORM transaction executor with SERIALIZABLE isolation_level and automatic retry.
        """
        return SessionTransactionExecutor(self.session_maker, tx_factory, engine=self.serializable_engine)



from bisect import bisect_left
from typing import Sequence, Any, Union


def seq_eq(a: Sequence, b: Sequence) -> bool:
    """
    Usable to check equality of items in a list and tuple. (or any custom sequence object)
    """
    if len(a) != len(b):
        return False
    return all(map(lambda x, y: x == y, a, b))


def seq_closest(seq: Sequence[Union[float, int]], value: Union[float, int]) -> Union[float, int]:
    """
    Return the closest numerical value in an ascending sorted sequence.
    In case of two values being equally close, choose the smaller.
    """
    if len(seq) == 0:
        raise ValueError("Can't find closest value in empty sequence")
    pos = bisect_left(seq, value)
    if pos == 0:
        return seq[0]
    if pos == len(seq):
        return seq[-1]
    before = seq[pos - 1]
    after = seq[pos]
    if after - value < value - before:
        return after
    return before


def seq_closest_index(seq: Sequence[Union[float, int]], value: Union[float, int]) -> int:
    """
    Return the index of the closest numerical value in an ascending sorted sequence.
    In case of two values being equally close, choose the earlier.
    """
    if len(seq) == 0:
        raise ValueError("Can't find closest value in empty sequence")
    pos = bisect_left(seq, value)
    if pos == 0:
        return 0
    if pos == len(seq):
        return len(seq) - 1
    before = seq[pos - 1]
    after = seq[pos]
    if after - value < value - before:
        return pos
    return pos - 1



from __future__ import annotations

import logging
import logging.handlers
from pathlib import Path
from typing import Optional, Union, TypeVar, Type

from tarka.logging.handler import (
    AbstractHandlerConfig,
    NamedHandlersAndHandlerConfigs,
    StreamHandlerConfig,
    RotatingFileHandlerConfig,
    FileHandlerConfig,
)

_HandlerT = TypeVar("_HandlerT", bound=logging.Handler)
_AbstractHandlerConfigT = TypeVar("_AbstractHandlerConfigT", bound=AbstractHandlerConfig)


class LoggerHandlerManager:
    """
    The logger objects are global. The root logger especially may get handlers from different contexts and there
    is no native solution to track them. Using this a library or application can easily add/remove/update their
    handlers on any logger safely, without affecting handlers unknown to them.
    """

    _INSTANCES = {}

    @staticmethod
    def instance(logger: Optional[logging.Logger] = None) -> LoggerHandlerManager:
        """
        Automatic global singleton instance support for convenience. Any number of LoggerHandlerManager may be
        used with any logger, but commonly only one is needed for the root.
        Loggers made by libraries for their own use usually won't need this, since those loggers are only
        utilized by that codebase most of the time.
        """
        if logger is None:
            logger = logging.getLogger()
        lhm = LoggerHandlerManager._INSTANCES.get(logger)
        if lhm is None:
            LoggerHandlerManager._INSTANCES[logger] = lhm = LoggerHandlerManager(logger)
        return lhm

    def __init__(self, logger: logging.Logger) -> None:
        self.logger = logger
        self.handlers: dict[str, logging.Handler] = {}  # that we manage

    def has(self, name: str) -> bool:
        return name in self.handlers

    def get(self, name: str) -> Optional[logging.Handler]:
        return self.handlers.get(name)

    def clear(self, close: bool = True) -> None:
        """
        Remove all managed handlers from the logger.
        """
        while self.handlers:
            _, handler = self.handlers.popitem()
            self.logger.removeHandler(handler)
            if close:
                handler.close()

    def remove(self, name: str, close: bool = True) -> Optional[logging.Logger]:
        """
        Remove a managed handler from the logger by name.
        Return the removed logger or None if no logger is tracked by name.
        """
        handler = self.handlers.pop(name, None)
        if handler is not None:
            self.logger.removeHandler(handler)
            if close:
                handler.close()
        return handler

    def add(self, name: str, handler: logging.Handler, close: bool = True) -> Optional[logging.Logger]:
        """
        Add a handler, replacing the previous one with the same name if any.
        Returns the previous handler that was replaced, None otherwise.
        """
        replaced_handler = self.remove(name, close=close)
        self.logger.addHandler(handler)
        self.handlers[name] = handler
        return replaced_handler

    # handler configs and setup convenience

    def remove_config(self, hc: AbstractHandlerConfig) -> None:
        self.remove(hc.name)

    def add_config(self, hc: Type[_AbstractHandlerConfigT[_HandlerT]]) -> _HandlerT:
        handler = hc.create()
        self.add(hc.name, handler)
        return handler

    def setup_handlers(self, nh_or_hc_list: NamedHandlersAndHandlerConfigs) -> None:
        # NOTE: When using the handler-config classes, one could specify the same handler multiple times with
        # different attributes for example, but the manager-name shall be different in that case!
        names_set_up = set()
        for nh_or_hc in nh_or_hc_list:
            if isinstance(nh_or_hc, AbstractHandlerConfig):
                name = nh_or_hc.name
                handler = nh_or_hc.create()
            elif (
                isinstance(nh_or_hc, tuple)
                and len(nh_or_hc) == 2
                and isinstance(nh_or_hc[0], str)
                and isinstance(nh_or_hc[1], logging.Handler)
            ):
                name, handler = nh_or_hc
            else:
                raise Exception(f"Invalid handler or config for setup: {nh_or_hc}")
            if name in names_set_up:
                raise Exception(
                    f"Already set up handler with name {name!r}: {self.get(name)} "
                    "Duplicate uses of the same handler config class need to have a unique name!"
                )
            self.add(name, handler)
            names_set_up.add(name)

    # specific, but generally useful handlers

    def remove_stderr_handler(self) -> None:
        self.remove(StreamHandlerConfig.DEFAULT_NAME)

    def add_stderr_handler(self) -> logging.StreamHandler:
        return self.add_config(StreamHandlerConfig())

    def remove_rotating_file_handler(self) -> None:
        self.remove(RotatingFileHandlerConfig.DEFAULT_NAME)

    def add_rotating_file_handler(
        self, log_directory: Union[str, Path], log_file_name: str = "rotating.log", max_bytes=4096000, backup_count=8
    ) -> logging.handlers.RotatingFileHandler:
        return self.add_config(RotatingFileHandlerConfig(log_directory, log_file_name, max_bytes, backup_count))

    def remove_file_handler(self) -> None:
        self.remove(FileHandlerConfig.DEFAULT_NAME)

    def add_file_handler(self, log_file: Union[str, Path]) -> logging.FileHandler:
        return self.add_config(FileHandlerConfig(log_file))



import logging
import logging.handlers
import sys
from io import TextIOWrapper
from pathlib import Path
from typing import Union, Sequence, Tuple, TypeVar, Generic, Optional

from tarka.logging.formatter import HandlerFormatterOrHandlerFormatterConfig, HandlerFormatterConfig

_HandlerT = TypeVar("_HandlerT", bound=logging.Handler)


class AbstractHandlerConfig(Generic[_HandlerT]):
    name: str

    def __init__(self, formatter: Optional[HandlerFormatterOrHandlerFormatterConfig]):
        self.formatter = formatter or HandlerFormatterConfig()

    def create(self) -> _HandlerT:
        handler = self._create()
        if self.formatter:
            HandlerFormatterConfig.apply(handler, self.formatter)
        return handler

    def _create(self) -> _HandlerT:
        raise NotImplementedError()


class RotatingFileHandlerConfig(AbstractHandlerConfig[logging.handlers.RotatingFileHandler]):
    DEFAULT_NAME = "ROTFILE"

    def __init__(
        self,
        log_path: Union[str, Path],
        log_file_name: str = None,
        max_bytes: int = 4096000,
        backup_count: int = 8,
        ensure_directory: bool = True,
        formatter: Optional[HandlerFormatterOrHandlerFormatterConfig] = None,
        name: str = DEFAULT_NAME,
    ):
        AbstractHandlerConfig.__init__(self, formatter)
        self.name = name
        log_path = Path(log_path)
        if log_file_name is None:
            self.log_directory = log_path.parent
            self.log_file_name = log_path.name
        else:
            self.log_directory = log_path
            self.log_file_name = log_file_name
        self.max_bytes = max_bytes
        self.backup_count = backup_count
        self.ensure_directory = ensure_directory

    def _create(self) -> logging.handlers.RotatingFileHandler:
        if self.ensure_directory:
            self.log_directory.mkdir(parents=True, exist_ok=True)
        return logging.handlers.RotatingFileHandler(
            str(self.log_directory / self.log_file_name),
            maxBytes=self.max_bytes,
            backupCount=self.backup_count,
        )


class FileHandlerConfig(AbstractHandlerConfig[logging.FileHandler]):
    DEFAULT_NAME = "ONEFILE"

    def __init__(
        self,
        log_path: Union[str, Path],
        log_file_name: str = None,
        ensure_directory: bool = True,
        formatter: Optional[HandlerFormatterOrHandlerFormatterConfig] = None,
        name: str = DEFAULT_NAME,
    ):
        AbstractHandlerConfig.__init__(self, formatter)
        self.name = name
        log_path = Path(log_path)
        if log_file_name is None:
            self.log_directory = log_path.parent
            self.log_file_name = log_path.name
        else:
            self.log_directory = log_path
            self.log_file_name = log_file_name
        self.ensure_directory = ensure_directory

    def _create(self) -> logging.FileHandler:
        if self.ensure_directory:
            self.log_directory.mkdir(parents=True, exist_ok=True)
        return logging.FileHandler(str(self.log_directory / self.log_file_name))


class StreamHandlerConfig(AbstractHandlerConfig[logging.StreamHandler]):
    DEFAULT_NAME = "STDERR"

    def __init__(
        self,
        stream: TextIOWrapper = None,
        formatter: Optional[HandlerFormatterOrHandlerFormatterConfig] = None,
        name: str = DEFAULT_NAME,
    ):
        AbstractHandlerConfig.__init__(self, formatter)
        self.name = name
        # NOTE: We acquire the reference to the stderr stream in the method body, so we'll see the current stream
        # if any post-import monkey-patching is being used.
        if stream is None:
            stream = sys.stderr
        self.stream = stream

    def _create(self) -> logging.StreamHandler:
        return logging.StreamHandler(self.stream)


NamedHandlersAndHandlerConfigs = Sequence[Union[Tuple[str, logging.Handler], AbstractHandlerConfig]]



"""
Handy functions to parse environment-variables to basic types.
"""

import os
from typing import Optional


def parse_bool_envvar(key: str, default: Optional[bool] = None) -> Optional[bool]:
    """
    Check if an environment variable is set with a "flag" value.

    :param key: Environment variable name to use.
    :param default: Value to return when the there is no environment variable with `key` name.
    :return: bool or `default`
    """
    if key not in os.environ:
        return default
    return os.environ[key].lower().strip() in ("1", "on", "yes", "true")


def parse_int_envvar(key: str, default: Optional[int] = None) -> Optional[int]:
    """
    Get an environment variable parsed as an integer.

    :param key: Environment variable name to use.
    :param default: Value to return when the there is no environment variable with `key` name.
    :return: int or `default`
    :raises ValueError when the environment variable's value can't be parsed as an integer.
    """
    if key not in os.environ:
        return default
    return int(os.environ[key])



import logging
import sys
import threading


def logging_excepthook(type_, value, traceback):
    logging.getLogger().critical("Unhandled exception", exc_info=(type_, value, traceback))


def logging_unraisablehook(unraisable):
    logging.getLogger().critical(
        "%s: %r",
        "Exception ignored in" if unraisable.err_msg is None else unraisable.err_msg,
        unraisable.object,
        exc_info=(unraisable.exc_type, unraisable.exc_value, unraisable.exc_traceback),
    )


def logging_threading_excepthook(args):
    if args.exc_type is not SystemExit:
        logging.getLogger().critical(
            "Unhandled exception in %r", args.thread, exc_info=(args.exc_type, args.exc_value, args.exc_traceback)
        )


def setup_logging_exception_hooks(
    excepthook: bool = True, unraisablehook: bool = True, threading_excepthook: bool = True
) -> None:
    if excepthook and not hasattr(sys, "original_excepthook"):
        sys.original_excepthook = sys.excepthook
        sys.excepthook = logging_excepthook
    if unraisablehook and not hasattr(sys, "original_unraisablehook"):
        sys.original_unraisablehook = sys.unraisablehook
        sys.unraisablehook = logging_unraisablehook
    if threading_excepthook and not hasattr(threading, "original_excepthook"):
        threading.original_excepthook = threading.excepthook
        threading.excepthook = logging_threading_excepthook



from typing import Iterable, Any, TypeVar, Generator, Iterator, Hashable, Union, Optional, Callable

from tarka.utility.sentinel import NamedObject

SENTINEL = NamedObject("sentinel")

_HashableT = TypeVar("_HashableT", bound=Hashable)
_AnyT = TypeVar("_AnyT", bound=Any)


def iter_unique(*item_iters: Iterable[_HashableT]) -> Generator[_HashableT, None, None]:
    """
    Yield unique(ly hashable) items for the iterables.
    """
    seen = set()
    for item_iter in item_iters:
        for item in item_iter:
            if item not in seen:
                seen.add(item)
                yield item


def select_eq_cmp_by_args(
    key: Optional[Callable[[_AnyT], Any]] = None, ascending: bool = True
) -> tuple[Callable[[_AnyT, _AnyT], bool], Callable[[_AnyT, _AnyT], bool], Callable[[_AnyT, _AnyT], bool]]:
    """
    Select the proper comparison lambdas to use for the properties.
    Return a lambda for testing equality, compare-not-equal and compare-or-equal.
    """
    if key is None:

        def _eq(i_: _AnyT, j_: _AnyT) -> bool:
            return i_ == j_

        if ascending:

            def _cnq(i_: _AnyT, j_: _AnyT) -> bool:
                return i_ < j_

            def _ceq(i_: _AnyT, j_: _AnyT) -> bool:
                return i_ <= j_

        else:

            def _cnq(i_: _AnyT, j_: _AnyT) -> bool:
                return i_ > j_

            def _ceq(i_: _AnyT, j_: _AnyT) -> bool:
                return i_ >= j_

    else:

        def _eq(i_: _AnyT, j_: _AnyT) -> bool:
            return key(i_) == key(j_)

        if ascending:

            def _cnq(i_: _AnyT, j_: _AnyT) -> bool:
                return key(i_) < key(j_)

            def _ceq(i_: _AnyT, j_: _AnyT) -> bool:
                return key(i_) <= key(j_)

        else:

            def _cnq(i_: _AnyT, j_: _AnyT) -> bool:
                return key(i_) > key(j_)

            def _ceq(i_: _AnyT, j_: _AnyT) -> bool:
                return key(i_) >= key(j_)

    return _eq, _cnq, _ceq


def iter_merge_zip(
    a: Iterable[_AnyT],
    b: Iterable[_AnyT],
    key: Optional[Callable[[_AnyT], Any]] = None,
    ascending: bool = True,
    s: NamedObject = SENTINEL,
) -> Generator[tuple[Union[_AnyT, NamedObject], Union[_AnyT, NamedObject]], None, None]:
    """
    Two-way merge-zip using the iterator interface.
    The iterators must yield their items sorted in in the order specified by `ascending`.
    Either of the yielded tuple values may be the sentinel which indicates miss, but not both.
    """
    _eq, _cnq, _ = select_eq_cmp_by_args(key, ascending)
    a = iter(a)
    b = iter(b)
    try:
        i = next(a)
    except StopIteration:
        i = s
    try:
        j = next(b)
    except StopIteration:
        j = s
    while i is not s and j is not s:
        if _eq(i, j):
            yield i, j
            try:
                i = next(a)
            except StopIteration:
                i = s
            try:
                j = next(b)
            except StopIteration:
                j = s
        elif _cnq(i, j):
            yield i, s
            try:
                i = next(a)
            except StopIteration:
                i = s
        else:
            yield s, j
            try:
                j = next(b)
            except StopIteration:
                j = s
    if i is not s:
        yield i, s
        while True:
            try:
                yield next(a), s
            except StopIteration:
                break
    if j is not s:
        yield s, j
        while True:
            try:
                yield s, next(b)
            except StopIteration:
                break


def iter_merge_two(
    a: Iterable[_AnyT],
    b: Iterable[_AnyT],
    key: Optional[Callable[[_AnyT], Any]] = None,
    ascending: bool = True,
    s: NamedObject = SENTINEL,
) -> Generator[_AnyT, None, None]:
    """
    Two-way merge using the iterator interface.
    The iterators must yield their items sorted in ascending order.
    """
    _, _, _ceq = select_eq_cmp_by_args(key, ascending)
    a = iter(a)
    b = iter(b)
    try:
        i = next(a)
    except StopIteration:
        i = s
    try:
        j = next(b)
    except StopIteration:
        j = s
    while i is not s and j is not s:
        if _ceq(i, j):
            yield i
            try:
                i = next(a)
            except StopIteration:
                i = s
        else:
            yield j
            try:
                j = next(b)
            except StopIteration:
                j = s
    if i is not s:
        yield i
        while True:
            try:
                yield next(a)
            except StopIteration:
                break
    if j is not s:
        yield j
        while True:
            try:
                yield next(b)
            except StopIteration:
                break


def iter_merge(
    *item_iters: Iterable[_AnyT],
    key: Optional[Callable[[_AnyT], Any]] = None,
    ascending: bool = True,
    s: NamedObject = SENTINEL
) -> Iterator[_AnyT]:
    """
    Multi merge using the iterator interface.
    The iterators must yield their items sorted in ascending order.
    """
    if not item_iters:
        return iter(())
    m = iter(item_iters[0])
    for i in range(1, len(item_iters)):
        m = iter_merge_two(m, iter(item_iters[i]), key=key, ascending=ascending, s=s)
    return m


def iter_merge_unique(
    *item_iters: Iterable[_AnyT], ascending: bool = True, s: NamedObject = SENTINEL
) -> Generator[_AnyT, None, None]:
    """
    Yield unique items for the iterables by merge-iterating over them.
    The iterators must yield their items sorted in ascending order.
    """
    m = iter_merge(*item_iters, ascending=ascending, s=s)
    try:
        last_item = next(m)
    except StopIteration:
        return  # all empty
    yield last_item
    try:
        while True:
            item = next(m)
            if item != last_item:
                yield item
                last_item = item
    except StopIteration:
        pass


def iter_all_same(*item_iters: Iterable[Any]) -> bool:
    """
    Check if all values equal in the iterables.
    """
    # select first value
    for i in range(len(item_iters)):
        it = iter(item_iters[i])
        for value in it:
            break
        else:
            continue  # empty iterable
        break
    else:
        return True  # all iterables are empty
    # continue iterable of first value
    for v in it:
        if value != v:
            return False
    i += 1
    while i < len(item_iters):
        for v in item_iters[i]:
            if value != v:
                return False
        i += 1
    return True



import errno
import os
import sys
import tempfile
from pathlib import Path
from typing import Optional, Callable, Tuple

from tarka.utility.file.name import split_extension


class ReserveFile:
    """
    Try and reserve a filename as is in by the given path, by creating it safely with EXCL flag. If that fails
    then the file name is to be adjusted with a random value to resolve conflicts.

    This is a variant of tempfile.mkstemp() that provides a behavioural override interface for adjusting the file-name
    and firstly tries to create the file without a random sequence. Additionally if the filename would be too long for
    the system, it will sensibly truncate characters.
    """

    def __init__(self, max_name_length: int = 255, fs_name_encode_length: Optional[Callable[[str], int]] = None):
        """
        :param max_name_length: This is only to optimize the initial truncation loop. Should match the maximum
            file-name length of the current platform.
        :param fs_name_encode_length: Customizable name-length calculator for the initial stripping round.
        """
        self.max_name_length = max_name_length
        if fs_name_encode_length is None:
            if sys.platform.startswith("linux"):
                self.fs_name_encode_length = self._fs_name_encode_utf8_bytes_length
            else:  # Assuming Windows, macOS or other systems that may have Unicode filename support.
                self.fs_name_encode_length = self._fs_name_encode_unicode_length
        else:
            self.fs_name_encode_length = fs_name_encode_length

        # mirrors how a binary file is created by mkstemp
        self.flags = os.O_RDWR | os.O_CREAT | os.O_EXCL
        if hasattr(os, "O_NOFOLLOW"):
            self.flags |= os.O_NOFOLLOW
        if hasattr(os, "O_BINARY"):
            self.flags |= os.O_BINARY

    def _fs_name_encode_unicode_length(self, name: str) -> int:
        """
        This is used for platforms that have support for Unicode filename representation on their primary filesystems.
        For these the unicode length shall be considered as the filename length as-is.
        """
        return len(name)

    def _fs_name_encode_utf8_bytes_length(self, name: str) -> int:
        """
        On Linux (and possibly other POSIX systems that originate from the 1970s) the filename length limit is measured
        as bytes, with a maximum length of 255. This limitation is so low-level in the kernel and all the file-systems
        as well, that it will not be enhanced soon.
        """
        return len(os.fsencode(name))

    def reserve(self, path_or_directory: Path, name: str = "", try_as_is: bool = True) -> str:
        """
        This is a convenience function that should be used if the file is reserved for later, and no operations are
        to be done with it right now.
        The reserved name is returned which could be different to the input name.
        The directory must exist.
        """
        fd, name = self.open_reserve(path_or_directory, name, try_as_is)
        os.close(fd)
        return name

    def open_reserve(self, path_or_directory: Path, name: str = "", try_as_is: bool = True) -> Tuple[int, str]:
        """
        Returns the open file-descriptor to the reserved file and its name, which could be different to the input name.
        The directory must exist.
        """
        if not name:
            prefix, suffix = self._adjust_name_direct(path_or_directory.name)
            directory = self._resolve(path_or_directory.parent)
        else:
            prefix, suffix = self._adjust_name_direct(name)
            directory = self._resolve(path_or_directory)
        if try_as_is:
            fd_name = self._reserve_adjust(self._mk_direct, suffix, prefix, directory, 0)
            if fd_name:
                return fd_name
        prefix, suffix = self._adjust_name_parts_temp(prefix, suffix)
        return self._reserve_adjust(self._mk_temp, suffix, prefix, directory, 8)  # mkstemp adds 8 char random sequence

    def _resolve(self, path: Path) -> Path:
        """
        Can be overridden if directory resolution errors need special care.
        """
        return path.resolve(strict=True)

    def _adjust_name_direct(self, name: str) -> Tuple[str, str]:
        """
        Can be overridden to customize the filename to be created at the direct reserve attempt.
        We need to make a the name split to prepare for inserting a random sequence in case of a conflict.
        """
        return split_extension(name)

    def _adjust_name_parts_temp(self, prefix: str, suffix: str) -> Tuple[str, str]:
        """
        Can be overridden to customize the filename to be created at the temp/random sequence attempts.
        The default implementation adds a separator character to the prefix, so the file name stay visually
        clear from random sequence. The suffix should start with a separator dot for the extension, so no
        change needed there.
        """
        if prefix.endswith("-"):
            return prefix, suffix
        return prefix + "-", suffix

    def _reserve_adjust(
        self,
        mk_fn: Callable[[str, str, Path], Optional[Tuple[int, str]]],
        suffix: str,
        prefix: str,
        directory: Path,
        mk_len: int,
    ) -> Optional[Tuple[int, str]]:
        while True:
            try:
                return mk_fn(suffix, prefix, directory)
            except OSError as e:
                if e.errno == errno.ENAMETOOLONG or (sys.platform.startswith("win") and e.errno == errno.EINVAL):
                    if len(prefix) + len(suffix) <= 1:
                        raise e  # no way to truncate more
                else:
                    raise e
            # adjust prefix/suffix to make name sorter
            pre_len = self.fs_name_encode_length(prefix)
            suf_len = self.fs_name_encode_length(suffix)
            while True:
                # NOTE: The underlying FS may enforce some normalization in addition to the encoding, so we actually
                # cannot accurately calculate how long the name would be in the representation of the FS. In addition
                # to that, different FS can have different file length support or may even change what the length
                # limit actually means in relation to the character encoding:
                #   - Windows counts the length limit in Unicode characters (as UTF-16 codepoints), having a limit of
                #     255 for NTFS, exFAT and FAT32, while having a lower limit on some other filesystems such as UDF.
                #   - macOS supports 255 Unicode characters with HFS+ similarly to Windows, but older or alternate
                #     filesystems may have more restrictive limits.
                #   - Linux counts the length limit as the encoded byte sequence up to 255 length in general.
                # So we truncate to sensible length in one go, then if that is still not acceptable we proceed by
                # a character each cycle with trial and error.
                if pre_len > suf_len:
                    prefix = prefix[:-1]
                    pre_len = self.fs_name_encode_length(prefix)
                elif suf_len > 0:
                    suffix = suffix[1:]
                    suf_len = self.fs_name_encode_length(suffix)
                if pre_len + suf_len + mk_len - 1 <= self.max_name_length:
                    break

    def _mk_direct(self, suffix: str, prefix: str, directory: Path) -> Optional[Tuple[int, str]]:
        # This is based on the internal mkstemp().
        name = prefix + suffix
        try:
            fd = os.open(directory / name, self.flags, 0o600)
        except FileExistsError:
            pass
        except PermissionError:
            if not sys.platform.startswith("win") or not directory.is_dir() or not os.access(directory, os.W_OK):
                raise
        else:
            return fd, name

    def _mk_temp(self, suffix: str, prefix: str, directory: Path) -> Tuple[int, str]:
        fd, temp_path = tempfile.mkstemp(suffix, prefix, directory)
        return fd, os.path.basename(temp_path)



import logging
from typing import Dict

from tarka.logging.handler import NamedHandlersAndHandlerConfigs, StreamHandlerConfig
from tarka.logging.hook import setup_logging_exception_hooks
from tarka.logging.level import LoggingLevelOrNamedLoggingLevelConfig, NamedLoggingLevelConfig
from tarka.logging.manager import LoggerHandlerManager
from tarka.logging.patch import TarkaLoggingPatcher


def init_tarka_logging() -> None:
    """
    Use this in a top level __init__ module, to ensure the Logger.trace is patched early for use.
    """
    TarkaLoggingPatcher.patch_custom_level(5, "TRACE")


def setup_basic_logging(
    root_handlers: NamedHandlersAndHandlerConfigs = None,
    logger_levels: Dict[str, LoggingLevelOrNamedLoggingLevelConfig] = None,
    setup_exc_hooks: bool = True,
    root_logger_level: LoggingLevelOrNamedLoggingLevelConfig = logging.WARNING,
) -> None:
    """
    Convenience function to setup loggers and handlers easily.
        - Setup handlers for the root logger. (because error logging shall be setup globally)
        - Setup levels for the specified loggers.

    A typical usage example:
        parser = argparse.ArgumentParser()
        parser.add_argument("--verbose", "-v", action="count", default=0)
        parser.add_argument("--silent", "-s", action="count", default=0)
        parser.add_argument("--log_file", help="path to write log to")
        parser.add_argument("--log_backups", type=int, default=16, help="how many log files to keep when rotating")
        args = parser.parse_args()
        setup_basic_logging(
            root_handlers=[RotatingFileHandlerConfig(args.log_file, backup_count=args.log_backups)],
            logger_levels={__name__: NamedLoggingLevelConfig(args.verbose, args.silent, default=logging.INFO)},
        )

    This should be used as early as possible in the application's lifespan.
    """
    if not root_handlers:  # Just select stderr for default.
        root_handlers = [StreamHandlerConfig()]

    root_lhm = LoggerHandlerManager.instance()
    root_lhm.setup_handlers(root_handlers)
    NamedLoggingLevelConfig.apply(root_lhm.logger, root_logger_level)

    for logger_name, level in (logger_levels or {}).items():
        NamedLoggingLevelConfig.apply(logging.getLogger(logger_name), level)

    if setup_exc_hooks:
        setup_logging_exception_hooks()



from __future__ import annotations

import logging
from typing import Union

from tarka.utility.algorithm.seq import seq_closest_index


def get_logging_levels_with_name(until_level: int = 101) -> list[int]:
    """
    It is astonishing that no public API exists to access the level-name mapping. The comments about levels
    say the NOTSET pseudo-level is the lower limit for user defined values. Since there should not be many
    more serious levels than FATAL, we iterate over all possible levels to see which has a name.
    We exclude the NOTSET from the returned collection.
    """
    levels = []
    for level in range(1, until_level):
        level_name = logging.getLevelName(level)
        if not level_name.startswith("Level "):
            levels.append(level)
    return levels


def calculate_named_logging_level(verbose: int = 0, silent: int = 0, default: int = logging.WARNING) -> int:
    """
    Calculate the most appropriate named logging level by verbosity and silence deltas relative to a default.
    """
    levels = get_logging_levels_with_name()
    default_index = seq_closest_index(levels, default)
    delta_index = silent - verbose
    return levels[min(max(0, default_index + delta_index), len(levels) - 1)]


class NamedLoggingLevelConfig:
    def __init__(self, verbose: int = 0, silent: int = 0, default: int = logging.WARNING):
        self.verbose = verbose
        self.silent = silent
        self.default = default

    @property
    def named_level(self) -> int:
        return calculate_named_logging_level(self.verbose, self.silent, self.default)

    @classmethod
    def apply(cls, logger: logging.Logger, level: LoggingLevelOrNamedLoggingLevelConfig) -> None:
        if isinstance(level, NamedLoggingLevelConfig):
            level = level.named_level
        logger.setLevel(level)


LoggingLevelOrNamedLoggingLevelConfig = Union[int, NamedLoggingLevelConfig]



from typing import Generator


def is_utf8_codepoint_start(b: int) -> bool:
    """
    Returns True if the byte-value could be the first byte of an UTF-8 codepoint.
    NOTE: Currently not used, more than 4 bytes sequence starts are also accepted. This should not be a problem,
    because they do not collide with the encoding of continuation bytes.
    :param b: assumed to be in range(256) like a value of a bytes sequence
    """
    return b < 0x80 or 0xC0 <= b


def is_utf8_codepoint_continuation(b: int) -> bool:
    """
    Returns True if the byte-value could be a continuation byte of an UTF-8 codepoint.
    :param b: a value of a bytes sequence
    """
    return 0x80 <= b < 0xC0


def iter_chunk_keep_utf8_codepoints(
    bs: bytes, chunk_size_max: int, longest_encoded_codepoint: int = 4
) -> Generator[bytes, None, None]:
    """
    Split the input bytes array into maximum-length chunks while respecting utf8 codepoint boundaries.
    """
    if chunk_size_max < longest_encoded_codepoint:
        raise RuntimeError("can't chunk utf8 shorter than longest encoded codepoint")
    i = 0
    j = chunk_size_max
    while j < len(bs):
        for _ in range(longest_encoded_codepoint - 1):
            if is_utf8_codepoint_start(bs[j]):
                break
            j -= 1
        yield bs[i:j]
        i = j
        j += chunk_size_max
    yield bs[i:]



import os
from io import StringIO

from alembic import command
from alembic.config import Config
from alembic.script import ScriptDirectory
from sqlalchemy import Connection


def get_alembic_config(alembic_dir: str, sqlalchemy_url: str = None) -> Config:
    alembic_cfg = Config(os.path.join(alembic_dir, "alembic.ini"))
    alembic_cfg.set_main_option("script_location", alembic_dir.replace("%", "%%"))
    if sqlalchemy_url:
        alembic_cfg.set_main_option("sqlalchemy.url", str(sqlalchemy_url).replace("%", "%%"))
    alembic_cfg.attributes["skip-logging-setup"] = True  # do not mess up the server logging
    return alembic_cfg


class NoHeadRevision(Exception):
    pass


class AlembicHelper:
    def __init__(self, config: Config):
        self.config = config

    def run(self, conn: Connection, alembic_command: str, *args: str) -> str:
        """
        Call the alembic CLI.
        """
        self.config.attributes["connection"] = conn
        self.config.stdout = StringIO()
        getattr(command, alembic_command)(self.config, *args)
        return self.config.stdout.getvalue()

    def run_strip_output(self, conn: Connection, alembic_command: str, *args: str) -> str:
        return self.run(conn, alembic_command, *args).strip()

    def get_head_revision(self) -> str:
        """
        Get current single head revision.
        """
        script = ScriptDirectory.from_config(self.config)
        revs = list(script.get_revisions("head"))
        if len(revs) != 1:
            raise NoHeadRevision()
        return revs[0].revision



import warnings
from contextlib import asynccontextmanager
from typing import AsyncContextManager, Union

from sqlalchemy import select, func
from sqlalchemy.ext.asyncio import AsyncSession, AsyncConnection


async def _pg_advisory_unlock(cos: Union[AsyncConnection, AsyncSession], *keys: int):
    unlocked = (await cos.execute(select(func.pg_advisory_unlock(*keys)))).scalar_one()
    if not unlocked:
        warnings.warn(
            "An inner scope has released the pg_try_advisory_lock() before the lock manager context did so explicitly. "
            "If such is expected, then call the lock manager with 'unlock=False' or restructure the transaction and"
            "locking to be strictly nested."
        )


@asynccontextmanager
async def pg_try_advisory_lock(
    cos: Union[AsyncConnection, AsyncSession], *keys: int, unlock: bool = True
) -> AsyncContextManager[bool]:
    """
    Acquire advisory session lock, yields False if the lock was not acquired.
    """
    locked = (await cos.execute(select(func.pg_try_advisory_lock(*keys)))).scalar_one()
    try:
        yield locked
    finally:
        if locked and unlock:
            await _pg_advisory_unlock(cos, *keys)


@asynccontextmanager
async def pg_advisory_lock(
    cos: Union[AsyncConnection, AsyncSession], *keys: int, unlock: bool = True
) -> AsyncContextManager:
    """
    Acquire advisory session lock, block until acquired.
    """
    (await cos.execute(select(func.pg_advisory_lock(*keys)))).scalar_one()
    try:
        yield
    finally:
        if unlock:
            await _pg_advisory_unlock(cos, *keys)



from typing import Callable, Union, TypeVar, Type, Optional, Iterator

JSONPrimitiveTypes = Union[str, int, float, list, dict, bool, None]
PythonPrimitiveTypes = Union[JSONPrimitiveTypes, bytes, tuple, set, frozenset]

TraverseType = TypeVar("TraverseType")
TraverseTypeMap = dict[Type[TraverseType], Callable[[TraverseType], Optional[Iterator[TraverseType]]]]


NATIVE_TRAVERSE: TraverseTypeMap = {
    list: lambda o: iter(o),
    tuple: lambda o: iter(o),
    set: lambda o: iter(o),
    frozenset: lambda o: iter(o),
    dict: lambda o: (x for k_v in o.items() for x in k_v),
}

# Only use this if all items in collection objects are comparable. (usually the same type).
SORTED_NAIVE_TRAVERSE: TraverseTypeMap = {
    **NATIVE_TRAVERSE,
    set: lambda o: iter(sorted(o)),
    frozenset: lambda o: iter(sorted(o)),
    dict: lambda o: iter(x for k in sorted(o.keys()) for x in (k, o[k])),
}

# Prefer to use the naive sorted traverse-map if possible for performance reasons.
SORTED_STR_TRAVERSE: TraverseTypeMap = {
    **NATIVE_TRAVERSE,
    set: lambda o: iter(sorted(o, key=lambda x: str(x))),
    frozenset: lambda o: iter(sorted(o, key=lambda x: str(x))),
    dict: lambda o: iter(x for k in sorted(o.keys(), key=lambda x: str(x)) for x in (k, o[k])),
}


class TraverseCycleError(RuntimeError):
    pass


def traverse(o: TraverseType, type_map: TraverseTypeMap = None) -> Iterator[TraverseType]:
    """
    Recursively traverse over an object structure by customizable traversal rules by object type.
    Each visited object is yielded and are directly followed by their children if any.
    The implementation uses a stack of iterator objects, it has no recursion depth limit.
    The objects that yield children are tracked in the traversal path to avoid cycles.
    """
    if type_map is None:
        # Use the native iteration order of primitive types if no traversal map is specified.
        type_map = NATIVE_TRAVERSE
    path: set[int] = set()
    stack: list[tuple[TraverseType, Iterator[TraverseType]]] = []
    yield o
    t = type_map.get(type(o))
    if t:
        if (o_id := id(o)) in path:
            raise TraverseCycleError()
        path.add(o_id)
        stack.append((o_id, t(o)))
    while stack:
        try:
            o = next(stack[-1][1])
        except StopIteration:
            path.remove(stack.pop()[0])
            continue
        yield o
        t = type_map.get(type(o))
        if t:
            if (o_id := id(o)) in path:
                raise TraverseCycleError()
            path.add(o_id)
            stack.append((o_id, t(o)))



import asyncio
import inspect
import random
import sqlite3
from typing import Any, Callable, Awaitable, Union, Type, Optional

from sqlalchemy.exc import DBAPIError, OperationalError
from sqlalchemy.ext.asyncio import AsyncSession, AsyncEngine, AsyncConnection


class AbstractRetryableTransaction:
    async def run(self) -> Any:
        raise NotImplementedError()


_MaybeRetryableTransaction = Union[AbstractRetryableTransaction, Awaitable[AbstractRetryableTransaction]]


class AbstractTransactionExecutor:
    def __init__(self, retry_delay: Optional[float] = None):
        self.retry_delay = retry_delay if retry_delay and retry_delay > 0 else 0

    async def run(self) -> Any:
        while True:
            try:
                return await self._run_tx()
            except OperationalError as e:
                # SQLite: This catches locking errors in both the rollback and WAL modes.
                if isinstance(e.orig, sqlite3.OperationalError) and e.orig.args == ("database is locked",):
                    pass
                else:
                    raise e
            except DBAPIError as e:
                # Postgresql: https://www.postgresql.org/docs/current/mvcc-serialization-failure-handling.html
                if e.orig and hasattr(e.orig, "sqlstate") and e.orig.sqlstate == "40001":
                    pass
                else:
                    raise e
            # TODO: implement more backend specific error handling (MariaDB, etc.)
            if self.retry_delay > 0:
                await asyncio.sleep(self.retry_delay * random.random())

    async def _run_tx(self) -> Any:
        raise NotImplementedError()

    async def _check_run_tx(self, tmp: _MaybeRetryableTransaction) -> Any:
        if inspect.isawaitable(tmp):
            tmp = await tmp
        if not isinstance(tmp, AbstractRetryableTransaction):
            raise Exception(f"not AbstractRetryableTransaction: {tmp}")
        return await tmp.run()


RetryableTransactionFactory = Callable[[AsyncConnection], _MaybeRetryableTransaction]


class TransactionExecutor(AbstractTransactionExecutor):

    def __init__(
        self, engine: AsyncEngine, tx_factory: RetryableTransactionFactory, retry_delay: Optional[float] = None
    ):
        AbstractTransactionExecutor.__init__(self, retry_delay)
        self.engine = engine
        self.tx_factory = tx_factory

    async def _run_tx(self) -> Any:
        async with self.engine.connect() as conn:
            async with conn.begin():
                return await self._check_run_tx(self.tx_factory(conn))


RetryableSessionTransactionFactory = Callable[[AsyncSession], _MaybeRetryableTransaction]


class SessionTransactionExecutor(AbstractTransactionExecutor):
    def __init__(
        self,
        session_maker: Type[AsyncSession],
        tx_factory: RetryableSessionTransactionFactory,
        engine: Optional[AsyncEngine] = None,
        retry_delay: Optional[float] = None,
    ):
        AbstractTransactionExecutor.__init__(self, retry_delay)
        self.session_maker = session_maker
        self.tx_factory = tx_factory
        self._session_maker_kwargs = {}
        if engine:
            self._session_maker_kwargs["bind"] = engine

    async def _run_tx(self) -> Any:
        async with self.session_maker(**self._session_maker_kwargs) as session:
            async with session.begin():
                return await self._check_run_tx(self.tx_factory(session))



import os
from datetime import datetime
from pathlib import Path
from typing import Union

from tarka.utility.time.format import PRETTY_TIMESTAMP


def get_log_file_path(log_file_path: Union[str, Path]) -> Path:
    """
    Evaluate the selected path and prepare it to be a logging target.
    Expands the path if it has the user's home (~).
    If the path does not end with a name of '.log' extension, treat it as a target directory and generate a filename.
    Create intermediate directories if they do not exist, which may raise an OSError for various reasons.
    """
    if not isinstance(log_file_path, Path):
        log_file_path = Path(log_file_path)
    log_file_path = log_file_path.expanduser()
    if log_file_path.name.endswith(".log"):
        log_parent = log_file_path.parent
    else:
        log_parent = log_file_path
        log_file_path = log_parent / f"{PRETTY_TIMESTAMP.date_time_micro(datetime.now())}_{os.urandom(4).hex()}.log"
    log_parent.mkdir(parents=True, exist_ok=True)
    return log_file_path



"""
The builtin logging facility works with global object instances which make them difficult to customize.
Yet this module shall add additional functionality to logging while keeping it completely compatible.
"""
import logging
from dataclasses import dataclass
from typing import Optional, Callable


@dataclass
class _HandlerMirrorHook:
    src: logging.Logger
    trg: logging.Logger

    def __call__(self, logger: logging.Logger, add: bool, handler: logging.Handler) -> None:
        if logger is self.src:
            if add:
                self.trg.addHandler(handler)
            else:
                self.trg.removeHandler(handler)


HandlerChangeHook = Callable[[logging.Logger, bool, logging.Handler], None]


class TarkaLoggingPatcher:
    HANDLER_TRACKING_PATCHED = False
    HANDLER_CHANGE_HOOKS: list[HandlerChangeHook] = []

    @classmethod
    def patch_handler_tracking(cls) -> None:
        """
        Add a callback hook to the Logger class for handler changes.
        """
        if cls.HANDLER_TRACKING_PATCHED:
            return

        def _logger_add_handler_override(self, handler) -> None:
            self.original_addHandler(handler)
            for hook in cls.HANDLER_CHANGE_HOOKS:
                hook(self, True, handler)

        def _logger_remove_handler_override(self, handler) -> None:
            self.original_removeHandler(handler)
            for hook in cls.HANDLER_CHANGE_HOOKS:
                hook(self, False, handler)

        logging.Logger.original_addHandler = logging.Logger.addHandler
        logging.Logger.addHandler = _logger_add_handler_override
        logging.Logger.original_removeHandler = logging.Logger.removeHandler
        logging.Logger.removeHandler = _logger_remove_handler_override
        cls.HANDLER_TRACKING_PATCHED = True

    @classmethod
    def hook_logger_mirror_handlers(cls, logger: logging.Logger, src_logger: Optional[logging.Logger] = None) -> None:
        """
        Copy the handlers to the logger from the source (root logger by default) and keep them in sync.
        This is useful when not propagating loggers need to be used, but they shall have the handlers of the root.
        """
        if src_logger is None:
            src_logger = logging.getLogger()
        if logger is src_logger:
            return
        cls.patch_handler_tracking()  # ensure hooks are patched in
        mirror_hook = _HandlerMirrorHook(src_logger, logger)
        if mirror_hook not in cls.HANDLER_CHANGE_HOOKS:
            cls.HANDLER_CHANGE_HOOKS.append(mirror_hook)
        cls.copy_logger_handlers(logger, src_logger)

    @classmethod
    def make_logger_shadow_root(cls, logger: logging.Logger) -> None:
        """
        Makes a logger (that is not root) stop propagating while mirroring the handlers from the root.
        This allows setting different levels for each library while all logging to the same handlers.
        """
        if logger is logging.getLogger():
            return
        logger.propagate = False
        cls.hook_logger_mirror_handlers(logger)

    @classmethod
    def copy_logger_handlers(cls, logger: logging.Logger, src_logger: Optional[logging.Logger] = None) -> None:
        """
        Copy the handlers to the logger from the source (root logger by default).
        """
        if src_logger is None:
            src_logger = logging.getLogger()
        for handler in src_logger.handlers:
            logger.addHandler(handler)

    @classmethod
    def patch_custom_level(cls, level: int, name: str) -> None:
        """
        Patch a custom log level to the Logger class.
        """
        assert name.isupper()
        if hasattr(logging.Logger, name.lower()):
            return  # assuming already patched

        logging.addLevelName(level, name)
        setattr(logging, name, level)

        def custom_log(self, message, *args, **kws):
            self._log(level, message, args, **kws)

        setattr(logging.Logger, name.lower(), custom_log)



import os
import shutil
from collections import defaultdict

from setuptools import setup, find_packages

HERE = os.path.abspath(os.path.dirname(__file__))
VERSION = "0.15.1"

with open(os.path.join(HERE, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

requirements = defaultdict(list)
for name in os.listdir(os.path.join(HERE, "requirements")):
    if name not in ("base.in", "asqla.in"):
        continue
    reqs = requirements[name.rpartition(".")[0]]
    with open(os.path.join(HERE, "requirements", name)) as f:
        for line in f:
            line = line.strip()
            if line.startswith("#") or line.startswith("-r"):
                continue
            reqs.append(line)
install_requirements = requirements.pop("base")


def _package_data_files(pkg: str, path: str = "", ignored_suffixes=(".py", ".pyc")):
    root = os.path.join(HERE, pkg)
    for parent, _, files in os.walk(os.path.join(root, path)):
        for file in files:
            if any(file.endswith(suffix) for suffix in ignored_suffixes):
                continue
            yield os.path.relpath(os.path.join(parent, file), root)


package_data = {"tarka": sorted(_package_data_files("tarka"))}
if package_data != {"tarka": []}:
    raise Exception(f"Package data has changed\n{str(package_data)}")

# Manually cleaning before build is required.
for p in [os.path.join(HERE, "build"), os.path.join(HERE, "dist"), os.path.join(HERE, "tarka.egg-info")]:
    if os.path.exists(p):
        shutil.rmtree(p)

setup(
    name="tarka",
    version=VERSION,
    description="Various things for common use, like a Swiss Army Knife.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Traktormaster/tarka",
    author="Nándor Mátravölgyi",
    author_email="nandor.matra@gmail.com",
    license="Apache 2.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
    ],
    packages=[p for p in find_packages(where=HERE, include=["tarka*"]) if not p.startswith("tarka_")],
    package_data=package_data,
    install_requires=install_requirements,
    extras_require=dict(requirements),
    python_requires=">=3.9",
)



import datetime

from sqlalchemy import TypeDecorator, DateTime


class UTCDateTime(TypeDecorator):
    impl = DateTime
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is not None:
            if not value.tzinfo:
                raise TypeError("tzinfo is required")
            value = value.astimezone(datetime.timezone.utc).replace(tzinfo=None)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = value.replace(tzinfo=datetime.timezone.utc)
        return value



"""
Tarka

:copyright: 2022-2023 Nándor Mátravölgyi
:license: Apache2, see LICENSE for more details.
"""

__copyright__ = "Copyright 2022-2023 Nándor Mátravölgyi"
__credits__ = ["Nándor Mátravölgyi"]
__version__ = "0.15.1"



"""
These are prepared for use in a server environment, where using local time is undesired.
All datetime objects returned are in UTC offset zero.
All timestamps are considered to be in UNIX epoch format.
"""
import re
import sys
import time
from datetime import datetime, timezone, timedelta
from uuid import UUID

ISO_DATETIME_RE = re.compile(
    r"^(\d{4})-(\d\d)-(\d\d)[T ](\d\d):(\d\d):(\d\d)(?:\.(\d{1,6}))?"
    r"(?:Z|([+-])(\d\d):?(\d\d):?(\d\d)?(?:\.(\d{1,6}))?)?$"
)


if sys.platform.startswith("win"):
    # NOTE: The C backend of fromtimestamp on Windows does not handle ranges before the epoch and farther in the
    # future, so... don't use it. Luckily the datetime facility in python can  bu used instead easily.
    def _dt_fromtimestamp(t, tz=None) -> datetime:
        return datetime.fromtimestamp(0, tz) + timedelta(seconds=t)

else:
    _dt_fromtimestamp = datetime.fromtimestamp


def utc_timestamp() -> float:
    """
    Current UTC time.
    In Python time.time() calls the time C function.
    On Linux, macOS, Windows the epoch is 1970-1-1, 00:00 UTC, and is timezone-independent.
    """
    return time.time()


def utc_timestamp_delta(offset_seconds: float) -> float:
    """
    Get the current timestamp with some offset.
    """
    return time.time() + offset_seconds


def utc_timestamp_from_datetime(dt: datetime) -> float:
    """
    This function flips the convention of assuming non-aware datetime to be in local time.
    On the server side if a remote request does not specify the timezone, assuming our timezone cannot be
    more correct than assuming the remote point to at least speak a common language, just badly.
    """
    if dt.utcoffset() is not None:
        return dt.timestamp()
    return dt.replace(tzinfo=timezone.utc).timestamp()


def utc_timestamp_from_uuid(uuid1: UUID) -> float:
    """
    The timestamp inside a UUID version 1 is not in UNIX epoch format, so we convert it to that.
    """
    return (uuid1.time - 0x01B21DD213814000) * 1e-07


def utc_datetime() -> datetime:
    """
    Current UTC time.
    Equivalent to calling: datetime.now(timezone.utc)
    """
    return _dt_fromtimestamp(time.time(), timezone.utc)


def utc_datetime_delta(offset_seconds: float) -> datetime:
    """
    Get the current timestamp with some offset.
    """
    return _dt_fromtimestamp(time.time() + offset_seconds, timezone.utc)


def utc_datetime_from_timestamp(utc_epoch_timestamp: float) -> datetime:
    """
    Convert the UTC timestamp to an UTC datetime.
    """
    return _dt_fromtimestamp(utc_epoch_timestamp, timezone.utc)


def utc_datetime_from_uuid(uuid1: UUID) -> datetime:
    """
    Convenience wrapper to get the timestamp of an UUID1 as datetime.
    """
    return _dt_fromtimestamp(utc_timestamp_from_uuid(uuid1), timezone.utc)


def utc_datetime_from_string(s: str) -> datetime:
    """
    Because datetime.strptime() can only handle a single static format, we regexp-process the input and construct
    the datetime object ourselves.
    """
    m = ISO_DATETIME_RE.match(s)
    if m is None:
        raise ValueError(f"invalid datetime for parse: {s}")
    groups = m.groups("0")
    # parse tz from timedelta
    if not (0 <= (tz_hours := int(groups[8])) < 24):
        raise ValueError(f"invalid tz hours for parse: {s}")
    if not (0 <= (tz_minutes := int(groups[9])) < 60):
        raise ValueError(f"invalid tz minutes for parse: {s}")
    if not (0 <= (tz_seconds := int(groups[10])) < 60):
        raise ValueError(f"invalid tz seconds for parse: {s}")
    tz_microseconds = int(groups[11] + "0" * (6 - len(groups[11])))
    tz_delta = timedelta(hours=tz_hours, minutes=tz_minutes, seconds=tz_seconds, microseconds=tz_microseconds)
    dt = datetime(
        int(groups[0]),
        int(groups[1]),
        int(groups[2]),
        int(groups[3]),
        int(groups[4]),
        int(groups[5]),
        int(groups[6] + "0" * (6 - len(groups[6]))) if groups[6] else 0,
        timezone.utc,
    )
    if groups[7] == "-":
        return dt + tz_delta
    return dt - tz_delta


def utc_seconds_until(utc_epoch_timestamp: float) -> float:
    return utc_epoch_timestamp - time.time()



class NamedObject(object):
    """
    Singleton named object generation and storage.
    These objects are ideal to be used as special values (sentinels) in algorithms where such are needed.
    """

    _MAP = dict()

    __slots__ = ("_name",)

    def __new__(cls, name):
        self = cls._MAP.get(name)
        if not self:
            cls._MAP[name] = self = object.__new__(cls)
        return self

    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    def __repr__(self):
        return f"NamedObject({self._name})"

    def __str__(self):
        return f"NamedObject({self._name})"



import os
import shutil
import tempfile
from pathlib import Path
from typing import Union


class SafeFile:
    """
    Provides a way to write and read a file on a path safely. Meaning that writing a new version of the file will
    never leave it in an undefined state. In other words, either the new file is successfully written or the old
    version is retained.

    This utility may be necessary on Windows, because the file moving/replacing (MoveFileEx) is not atomic.

    Write new versions of a file safely by utilizing a temporary file and keeping a backup.
    Read the file and fallback to the backup if necessary.

    Not thread-safe!
    """

    __slots__ = ("normal", "backup")

    def __init__(self, path: Union[str, Path], backup_suffix: str = "bak"):
        if not isinstance(path, Path):
            path = Path(path)
        self.normal = path
        self.backup = path.with_name(f"{path.name}.{backup_suffix}")

    def read(self) -> bytes:
        """
        Read the normal path or fall back to the backup.
        """
        try:
            return self.normal.read_bytes()
        except OSError:
            return self.backup.read_bytes()

    def write(self, data: bytes) -> None:
        """
        Write a new version of the file, manage the backup automatically.
        """
        fd, temp_path = tempfile.mkstemp(dir=self.normal.parent, text=False)
        try:
            with os.fdopen(fd, "wb") as f:
                f.write(data)  # Write new version to temp file.
            if self.normal.exists():  # Backup old version if it exists.
                # Removing the backup is necessary to avoid consistency issues made by overwriting.
                # Removing the backup is safe, because if the normal path exists it must be a valid version,
                # as it only ever gets replaced (moved) over. In other words, the backup is never newer than
                # the normal path when that exists.
                self.backup.unlink(missing_ok=True)  # Remove old backup first.
                shutil.copy2(self.normal, self.backup)  # Copy old version to backup.
            # The implementation of replace is atomic on POSIX compatible systems, so the backup feature will
            # never be utilized. However, on Windows the move may be a delete&move on the FS metadata level and
            # thus the old version might be lost. By having a backup of the last version this issue is handled.
            os.replace(temp_path, self.normal)  # Replace old version with temp (new).
        finally:
            try:
                os.unlink(temp_path)  # do not leave the temp file for whatever reason.
            except OSError:
                pass
